#!/usr/bin/env python3
'''
mock_server.py - this file is part of S3QL (http://s3ql.googlecode.com)

Copyright (C) Nikolaus Rath <Nikolaus@rath.org>

This program can be distributed under the terms of the GNU GPLv3.
'''

from http.server import BaseHTTPRequestHandler, _quote_html
import re
import threading
import socketserver
import logging
from urlparse import urlsplit, parse_qs
from xml.sax.saxutils import escape as xml_escape

log = logging.getLogger(__name__)

def start_server(host, port):
    server = HTTPServer((host, port))
    # Initializes .server_name and .server_port
    server.thread = threading.Thread(target=server.serve_forever)
    server.thread.start()
    return server

# Shutdown with: 
# server.shutdown()
# server.server_close()

class HTTPServer(socketserver.ThreadingTCPServer):

    def __init__(self, server_address):
        super().__init__(server_address, RequestHandler, )
        self.data = dict()
        self.metadata = dict()
        
class RequestHandler(BaseHTTPRequestHandler):

    server_version = "MockHTTP"
    protocol_version = 'HTTP/1.1'
    meta_header_re = re.compile(r'X-AMZ-Meta-([[:alnum:]-_]+)$',
                                re.IGNORECASE)
    xml_ns_prefix = '{http://s3.amazonaws.com/doc/2006-03-01/}'
    
    def log_message(self, format, *args):
        log.debug(format, *args)

    def handle_DELETE(self):
        q = urlsplit(self.path)
        try:
            del self.server.data[q.path]
            del self.server.metadata[q.path]
        except KeyError:
            self.send_error(404)
            return
        else:
            self.send_response(204)
            self.end_headers()

    def handle_PUT(self):
        encoding = self.headers['Content-Encoding']
        len_ = self.headers['Content-Length']
        if not len_ or (encoding and encoding != 'identity'):
            self.send_error(415)
            return
        len_ = int(len_)

        q = urlsplit(self.path)
        meta = dict()
        for (name, value) in self.headers.items:
            hit = self.meta_header_re.match(name)
            if hit:
                meta[hit.group(1)] = value

        src = headers.get('x-amz-copy-source')
        if src:
            hit = re.match('^/([[:alnum:]]+)/(.+)$', src)
            if not hit:
                
        # TODO: etag
                
        self.server.metadata[q.path] = meta
        self.server.data[q.path] = self.rfile.read(len_)
        self.send_response(201)
        self.end_headers()

    def handle_expect_100(self):
        if self.command == 'PUT':
            encoding = self.headers['Content-Encoding']
            len_ = self.headers['Content-Length']
            if not len_ or (encoding and encoding != 'identity'):
                self.send_error(415)
                return False
            
        self.send_response_only(100)
        self.end_headers()
        return True

    def do_GET(self):
        q = urlsplit(self.path)

        if q.path == '/':
            return self.do_list(q)

        try:
            data = self.server.data[q.path]
            meta = self.server.metadata[q.path]
        except KeyError:
            self.send_error(404)
            return
        
        self.send_response(200)
        self.send_header("Content-Type", 'application/octet-stream')
        self.send_header("Content-Length", str(len(data)))
        for tup in meta.items():
            self.send_header('X-AMZ-Meta-%s: %s' % tup)
        self.end_headers()
        self.wfile.write(data)

    def do_list(self, q):
        qs = parse_qs(q.query)
        marker = qs.get('marker', None)
        prefix = qs.get('prefix', '')
        max_keys = qs.get('max_keys', 1000)

        resp = ['<?xml version="1.0" encoding="UTF-8"?>',
                '<ListBucketResult xmlns="http://s3.amazonaws.com/doc/2006-03-01/">',
                '<MaxKeys>%d</MaxKeys>' % max_keys,
                '<IsTruncated>false</IsTruncated>' ]
                
        count = 0
        for key in self.server.data:
            if not key.startswith(prefix):
                continue
            if marker and key <= marker:
                continue
            resp.append('<Contents><Key>%s</Key></Contents>' % xml_escape(key))
            count += 1
            if count == max_keys:
                resp[3] = '<IsTruncated>true</IsTruncated>'
                break

        resp.append('</ListBucketResult>')
        body = '\n'.join(resp).encode()
            
        self.send_response(200)
        self.send_header("Content-Type", 'text/xml')
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)
        
        
    def do_HEAD(self):
        q = urlsplit(self.path)
        try:
            meta = self.server.metadata[q.path]
            data = self.server.data[q.path]
        except KeyError:
            self.send_error(404)
            return
        
        self.send_response(200)
        self.send_header("Content-Type", 'application/octet-stream')
        self.send_header("Content-Length", str(len(data)))
        for tup in meta.items():
            self.send_header('X-AMZ-Meta-%s: %s' % tup)
        self.end_headers()

    def send_error(self, code, message=None):
        # Overwritten to not close connection on errors and provide
        # content-length
        try:
            shortmsg, longmsg = self.responses[code]
        except KeyError:
            shortmsg, longmsg = '???', '???'
        if message is None:
            message = shortmsg
        explain = longmsg
        self.log_error("code %d, message %s", code, message)
        # using _quote_html to prevent Cross Site Scripting attacks (see bug #1100201)
        content = (self.error_message_format % {'code': code, 'message': _quote_html(message),
                                               'explain': explain}).encode('utf-8', 'replace')
        self.send_response(code, message)
        self.send_header("Content-Type", self.error_content_type)
        self.send_header("Content-Length", str(len(content)))
        self.end_headers()
        if self.command != 'HEAD' and code >= 200 and code not in (204, 304):
            self.wfile.write(content)
        
